//
//  NSBezierPath-NoodleExtensions.m
//  Hazel
//
//  Created by Paul Kim on 2/18/06.
//  Copyright 2006 Paul Kim d/b/a Noodlesoft. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in 
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
// IN THE SOFTWARE.


#import "NSBezierPath-NoodleGradient.h"

static BOOL		_debug = NO;

typedef struct ColorVector
{
	float	r;
	float	g;
	float	b;
	float	a;
} ColorVector;

static void evaluateLinearGradient(void *info, const float *in, float *out)
{
	ColorVector	*colors;
	
	colors = info;

	*out++ = colors[0].r + *in * colors[1].r;
	*out++ = colors[0].g + *in * colors[1].g;
	*out++ = colors[0].b + *in * colors[1].b;
	*out = colors[0].a + *in * colors[1].a;
}

@implementation NSBezierPath (NoodleGradient)

- (void)gradientFillAtAngle:(float)angle startColor:(NSColor *)startColor endColor:(NSColor *)endColor
{
	NSRect				bounds;
	NSPoint				startPoint, endPoint;
	float				angleTolerance;

	bounds = [self bounds];
	
	// Calculate the angle that it would take to cause an arc of half a point
	// in the current coordinate system. We take the longer of the two sides of the bounds.
	angleTolerance = asinf(0.5 / MAX(NSWidth(bounds), NSHeight(bounds)));
	
	// Normalize the angle
	angle = angle - ((int)(angle / 360) * 360);

	if (fabsf(angle - 0.0) < angleTolerance)
	{
		startPoint.x = NSMinX(bounds);
		startPoint.y = NSMinY(bounds);
		endPoint.x = NSMaxX(bounds);
		endPoint.y = startPoint.y;
	}
	else if (fabsf(angle - 90.0) < angleTolerance)
	{
		startPoint.x = NSMinX(bounds);
		startPoint.y = NSMinY(bounds);
		endPoint.x = startPoint.x;
		endPoint.y = NSMaxY(bounds);
	}
	else if (fabsf(angle - 180.0) < angleTolerance)
	{
		startPoint.x = NSMaxX(bounds);
		startPoint.y = NSMinY(bounds);
		endPoint.x = NSMinX(bounds);
		endPoint.y = startPoint.y;
	}
	else if (fabsf(angle - 270.0) < angleTolerance)
	{
		startPoint.x = NSMinX(bounds);
		startPoint.y = NSMaxY(bounds);
		endPoint.x = startPoint.x;
		endPoint.y = NSMinY(bounds);
	}
	else
	{
		NSAffineTransform	*transform;
		NSBezierPath		*transformedPath;
		NSRect				realBounds;

		bounds = [self bounds];

		// We are creating this transform to transform the path, not the actual
		// coordinate system. We are simulating the path in the transformed space
		// to get the correct bounding rect. Doing this via the coordinate system
		// seems to return the bounding rect of the bounding rect of where the path
		// was added.
		transform = [NSAffineTransform transform];
		[transform rotateByDegrees:-angle];

		// Copy the path (we don't want to alter the original) and transform to
		// get it's bounding rect in the transformed space
		transformedPath = [transform transformBezierPath:self];
		
		// Get the real bounding rect.
		realBounds = [transformedPath bounds];

		// Set start and end points to the bottom edge of the bounding rect
		startPoint = realBounds.origin;
		endPoint.x = NSMaxX(realBounds);
		endPoint.y = startPoint.y;
		
		// Reverse the transform to convert the points back to regular space
		[transform invert];

		// Convert the points to the original coordinate system.
		startPoint = [transform transformPoint:startPoint];
		endPoint = [transform transformPoint:endPoint];

		if (_debug)
		{
			// If debugging, draw the rotated bounding box in red
			[[NSGraphicsContext currentContext] saveGraphicsState];
			
			// Use pre-existing transform (it has been inverted)
			[transform concat];
			
			[[NSColor redColor] set];
			[NSBezierPath strokeRect:realBounds];
			
			[[NSGraphicsContext currentContext] restoreGraphicsState];
		}
	}

	[self gradientFillWithStartPoint:startPoint endPoint:endPoint startColor:startColor endColor:endColor];
}
	
- (void)gradientFillWithStartPoint:(NSPoint)startPoint endPoint:(NSPoint)endPoint startColor:(NSColor *)startColor endColor:(NSColor *)endColor
{
	static const float						gradientDomain[] = { 0.0, 1.0 };
	static const float						colorRanges[] = { 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0 };
	static const CGFunctionCallbacks		gradientCallbacks = { 0, &evaluateLinearGradient, NULL };	
	
	CGFunctionRef		function;
	ColorVector			colors[2];
	CGColorSpaceRef		colorspace;
	CGShadingRef		shading;
	CGContextRef		context;
	

	context = (CGContextRef)[[NSGraphicsContext currentContext] graphicsPort];
	
	if (![[startColor colorSpaceName] isEqual:NSCalibratedRGBColorSpace])
	{
		startColor = [startColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
	}
	if (![[endColor colorSpaceName] isEqual:NSCalibratedRGBColorSpace])
	{
		endColor = [endColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
	}
	
	[startColor getRed:&colors[0].r green:&colors[0].g blue:&colors[0].b alpha:&colors[0].a];
	[endColor getRed:&colors[1].r green:&colors[1].g blue:&colors[1].b alpha:&colors[1].a];

	// Pre-calculate difference
	colors[1].r -= colors[0].r;
	colors[1].g -= colors[0].g;
	colors[1].b -= colors[0].b;
	colors[1].a -= colors[0].a;
	
	function = CGFunctionCreate(colors, 1, gradientDomain, 4, colorRanges, &gradientCallbacks);
	
	if (function == NULL)
	{
		NSLog(@"Could not create gradient function");
		return;
	}
	
	colorspace = CGColorSpaceCreateDeviceRGB();
	shading = CGShadingCreateAxial(colorspace, *(CGPoint *)&startPoint, *(CGPoint *)&endPoint, function, true, true);

	CGContextSaveGState(context);
	[self addClip];
	
	CGContextDrawShading(context, shading);
	CGContextRestoreGState(context);

	CGFunctionRelease(function);	
	CGShadingRelease(shading);
	CGColorSpaceRelease(colorspace);
}

@end
